#include <stdio.h>
#include <unistd.h>
#include <vga.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include <sys/signal.h>
#include <stdlib.h>
#include "libvga.h"

/*
#define HAS_CARD_INFO
*/

static int client_memfd;
static int svgalib_linear_mem_size;
static unsigned char *svgalib_linear_mem_base, *client_linear_mem_virtual;
unsigned char *graph_mem;

static unsigned char *input_space, *output_space;

static int shmid, semid;
static struct sembuf semarray[2];
int child_id;
int client_currentmode = 0, client_lastmode = 0, client_color, 
	client_displaystart;
	

void wake_up_server(void)
{
	/* This function also puts us to sleep */
	semarray[0] . sem_num = 1;
	semarray[0] . sem_op =  1;
	semarray[0] . sem_flg = 0;
	semarray[1] . sem_num = 0;
	semarray[1] . sem_op = -1;
	semarray[1] . sem_flg = 0;
	if(semop(semid, semarray, 1)){
		fprintf(stderr, "client: could not perform semaphore "
				"operation\n");
		exit(1);
	}
	if(semop(semid, semarray + 1, 1)){
		fprintf(stderr, "client: could not perform semaphore "
				"operation\n");
		exit(1);
	}
}

void client_atexit(void){
#if !defined(HAS_CARD_INFO)
	*((int *) output_space) = 27;
#else
	*((int *) output_space) = 28;
#endif
	semarray[0] . sem_num = 1;
	semarray[0] . sem_op =  1;
	semarray[0] . sem_flg = 0;
	if(semop(semid, semarray, 1)){
		fprintf(stderr, "client: could not perform semaphore "
				"operation\n");
		exit(1);
	}
	if(semctl(semid, 0, IPC_RMID, 0) < 0){
		fprintf(stderr, "client: unable to remove semaphores 4754h\n");
	}
	if(shmctl(shmid, IPC_RMID, 0) < 0)
		fprintf(stderr, "client: unable to mark shared memory"
			       " segment as to be destroyed\n");
}

#define CODE_0_0(name, code) \
void name(void) \
{ \
	((int *) output_space)[0] = code; \
	wake_up_server(); \
}

#define CODE_4_4(name, code) \
int name(int x0) \
{ \
	((int *) output_space)[0] = code; \
	((int *) output_space)[1] = x0; \
	wake_up_server(); \
	return *((int *) input_space); \
}

#define CODE_4_0(name, code) \
int name(void) \
{ \
	((int *) output_space)[0] = code; \
	wake_up_server(); \
	return *((int *) input_space); \
}

#define CODE_0_4(name, code) \
void name(int x0) \
{ \
	((int *) output_space)[0] = code; \
	((int *) output_space)[1] = x0; \
	wake_up_server(); \
}	

#define CODE_4_16(name, code) \
int name(int x0, int x1, int x2, int x3) \
{ \
	((int *) output_space)[0] = code; \
	((int *) output_space)[1] = x0; \
	((int *) output_space)[2] = x1; \
	((int *) output_space)[3] = x2; \
	((int *) output_space)[4] = x3; \
	wake_up_server(); \
	return *((int *) input_space); \
}

vga_modeinfo client_modeinfo;
int (*vga_drawpixelfunction)(int, int);
int vga_drawpixel8(int x, int y)
{
	char *p;
	p = graph_mem + client_modeinfo . linewidth * y + 
		x;
	*p = client_color;
}
int vga_drawpixel16(int x, int y)
{
	char *p;
	p = graph_mem + client_modeinfo . linewidth * y + 
		x * 2;
	*((short *)p) = client_color;
}

int vga_drawpixel24(int x, int y)
{
	char *p;
	p = graph_mem + client_modeinfo . linewidth * y + 
		x * 3;
	*((short *)p) = (short)client_color;
	*(p + 2) = client_color >> 16;
}
void (*vga_drawpixeladdress)(void *);

void vga_drawpixeladdress8( void *p)
{
	*((char *) p) = client_color;
}

void vga_drawpixeladdress16(void *p)
{
	*((short *) p) = client_color;
}

void vga_drawpixeladdress24(void *p)
{
	*((short *) p) = client_color;
	*(((char *) p ) + 2) = client_color >> 16;
}

int vga_setmode(int x0) 
{
	/* Keeping from setting VGA modes because backward compatibility
	   is difficult for video card manufacturers.
	 */
	if(x0 && (x0 < 10)){
		fprintf(stderr, "client: vga modes not available.\n");
		exit(1);
	}
	client_lastmode = client_currentmode;
	client_currentmode = x0;	
	((int *) output_space)[0] = 0; 
	((int *) output_space)[1] = x0; 
	wake_up_server(); 
	client_modeinfo = *((vga_modeinfo *)(input_space + sizeof(int)));
	switch(client_modeinfo . bytesperpixel){
		case 1:	vga_drawpixelfunction = &vga_drawpixel8;
			vga_drawpixeladdress = &vga_drawpixeladdress8;
			client_color = 1;
			break;
		case 2:	vga_drawpixelfunction = &vga_drawpixel16;
			vga_drawpixeladdress = &vga_drawpixeladdress16;
			client_color = 0x0410;
			break;
		case 3:	vga_drawpixelfunction = &vga_drawpixel24;
			vga_drawpixeladdress = &vga_drawpixeladdress24;
			client_color = 0x008080;
			break;
	}
	return *((int *) input_space); 
}

CODE_4_4(vga_hasmode, 1);

int vga_clear(void)
{
	memset(client_linear_mem_virtual + client_displaystart, 0, 
		client_modeinfo . height * client_modeinfo . linewidth *
		client_modeinfo . bytesperpixel);
	return 0;
}
CODE_4_0(vga_flip, 2);

int vga_getxdim(void)
{
	return client_modeinfo . width;
}

int vga_ydim(void)
{
	return client_modeinfo . height;
}

int vga_getcolors(void)
{
	return client_modeinfo . colors;
}

CODE_4_16(vga_setpalette, 3);
int vga_getpalette(int index, int *red, int *green, int *blue)
{
	((int *) input_space)[0] = 4;
	((int *) input_space)[1] = index;
	wake_up_server();
	*red = ((int *) output_space)[1];
	*green = ((int *) output_space)[2];
	*blue = ((int *) output_space)[3];
	return ((int *) output_space)[0];
}

int vga_setpalvec(int start, int num, int *pal)
{
	((int *) input_space)[0] = 5;
	((int *) input_space)[1] = num;
	memcpy(input_space + 2 * sizeof(int), pal, num * sizeof(int));
	wake_up_server();
	return ((int *) output_space)[0];
}

int vga_getpalvec(int start, int num, int *pal)
{
	((int *) input_space)[0] = 6;
	((int *) input_space)[1] = num;
	wake_up_server();
	memcpy(pal, output_space + sizeof(int), num * sizeof(int));
	return ((int *) output_space)[0];
}
CODE_4_0(vga_screenoff, 7);
CODE_4_0(vga_screenon, 8);

int vga_setcolor(int color)
{
	client_color = color;
	return 0;
}


int vga_drawpixel(int x, int y)
{
	(*vga_drawpixelfunction)(x, y);
}


int vga_drawline(int x1, int y1, int x2, int y2)
{
	int dx, dy, togo, incr, num, denum, aincr;
	char *addr;

	dx = x2 - x1;
	dy = y2 - y1;
	if(abs(dx) > abs(dy)){
		denum = togo = abs(dx);
		if(x1 < x2) {
			addr = graph_mem + client_modeinfo . linewidth * y1 +
				x1 * client_modeinfo . bytesperpixel;
			if(y2 < y1) {
				aincr = - client_modeinfo . linewidth;
				incr = y1 - y2;
			}
			else {
				aincr = client_modeinfo . linewidth;
				incr = y2 - y1;
			}
		} else {
			addr = graph_mem + client_modeinfo . linewidth * y2 +
				x2 * client_modeinfo . bytesperpixel;
			if(y1 < y2) {
				aincr = - client_modeinfo . linewidth;
				incr = y2 - y1;
			} else {
				aincr = client_modeinfo . linewidth;
				incr = y1 - y2;
			}
		}
		num = denum / 2;
		do {
			(*vga_drawpixeladdress)(addr);
			addr += client_modeinfo . bytesperpixel;
			num += incr;
			if(num >= denum){
				addr += aincr;
				num -=denum;
			}
		} while(--togo);
	} else {
		denum = togo = abs(dy);
		if(y1 < y2) {
			addr = graph_mem + client_modeinfo . linewidth * y1 +
				x1 * client_modeinfo . bytesperpixel;
			if(x2 < x1) {
				aincr = -client_modeinfo . bytesperpixel;
				incr = x1 - x2;
			} else {
				aincr = +client_modeinfo . bytesperpixel;
				incr = x2 - x1;
			}
		} else {
			addr = graph_mem + client_modeinfo . linewidth * y2 +
				x2 * client_modeinfo . bytesperpixel;
			if(x1 < x2) {
				aincr = -client_modeinfo . bytesperpixel;
				incr = x2 - x1;
			} else {
				aincr = +client_modeinfo . bytesperpixel;
				incr = x1 - x2;
			}
		}
		num = denum / 2;
		do {
			(*vga_drawpixeladdress)(addr);
			addr += client_modeinfo . linewidth;
			num += incr;
			if(num >= denum){
				addr += aincr;
				num -=denum;
			}
		} while(--togo);
	}
}

int vga_drawscansegment(unsigned char *colors, int x, int y, int length)
{
	char *p;
	p = graph_mem + client_modeinfo . linewidth * y + 
		x * client_modeinfo . bytesperpixel;
	memcpy(p, colors, length);
}
int vga_drawscanline(int line, unsigned char *colors)
{
	vga_drawscansegment(colors, 0, line, client_modeinfo . width * 
			client_modeinfo . bytesperpixel);
}

int vga_getpixel(int x, int y)
{
	char *p;
	int i;
	
	p = graph_mem + client_modeinfo . linewidth * y; 
	switch(client_modeinfo . bytesperpixel){
	case 1:	p += x;
		return *p;
	case 2:	p += 2 * x;
		return *((short *) p);
	case 3:	p += 3 * x;
		return *((int *) p) & 0xffffff;
	}
}

int vga_getscansegment(unsigned char *colors, int x, int y, int length)
{
	char *p;
	p = graph_mem + client_modeinfo . linewidth * y + 
		x * client_modeinfo . bytesperpixel;
	memcpy(colors, p, length);
}
CODE_4_0(vga_dumpregs, 9);

vga_modeinfo *vga_getmodeinfo(int mode)
{
	static vga_modeinfo vmi;
	
	*((int *) output_space) = 10;
	wake_up_server();
	vmi = *((vga_modeinfo *)input_space);
	return &vmi;
}

#if defined(HAS_CARD_INFO)

vga_cardinfo *vga_getcardinfo(void)
{
	static vga_cardinfo vci;
	*((int *) output_space) = 27;
	wake_up_server();
	vci = *((vga_cardinfo *) input_space);
	return &vci;
}

#endif

CODE_4_0(vga_getdefaultmode, 11);
int vga_getcurrentmode(void)
{
	return client_currentmode;
}

int vga_getcurrentchipset(void)
{
	return client_modeinfo . chiptype;
}

char *vga_getmodename(int mode)
{
	char *c;
	((int *) output_space)[0] = 12;
	((int *) output_space)[1] = mode;
	c = malloc(strlen(input_space) + 1);
	strcpy(c, input_space);
	return c;
}

int vga_getmodenumber(char *name)
{
	((int *) output_space)[0] = 13;
	strcpy(output_space + sizeof(int), name);
	return ((int *) input_space)[0];
}

int vga_getoptmode(int x, int y, int colors, int bytesperpixel, int c)
{
	((int *) output_space)[0] = 14;
	((int *) output_space)[1] = x;
	((int *) output_space)[2] = y;
	((int *) output_space)[3] = colors;
	((int *) output_space)[4] = bytesperpixel;
	((int *) output_space)[5] = c;
	wake_up_server();
	return ((int *) input_space)[0];
}

int vga_accel(unsigned operation, ... )
{
	/* Acceleration functions not available :-( */
	return -1;
}

/* We invoke the old interrupt handler after setting text mode */
/* We catch all signals that cause an exit by default (aka almost all) */
static char sig2catch[] ={
	SIGHUP, SIGINT, SIGQUIT, SIGILL,
	SIGTRAP, SIGIOT, SIGBUS, SIGFPE,
	SIGSEGV, SIGPIPE, SIGALRM, SIGTERM,
	SIGXCPU, SIGXFSZ, SIGVTALRM, SIGPWR
};

static struct sigaction old_signal_handler[sizeof(sig2catch)];
static void signal_handler(int v)
{
	int i;

	printf("vga client: signal %d received.\n", v);
	client_atexit();
    	for (i = 0; i < sizeof(sig2catch); i++)
		if (sig2catch[i] == v) {
			sigaction(v, old_signal_handler + i, 0);
			raise(v);
			break;
		}
}

/* mode = 	0:	vga_init
   		1:	vga_initf
		2:	vga_simple_init
 */
int client_init(int mode, int flags){
	unsigned char c, carr[40];
	int i;
	struct sigaction siga;

	/* Creating a shared memory segment */
	if((shmid = shmget(0x4754, 8000, IPC_CREAT | 384)) < 0){
		fprintf(stderr, "client:\tcould not create shared memory "
				"segment with key 4754h.\n");
		exit(1);
	}
	/* Creating two semaphores for process synchronization */
	if((semid = semget(0x4754, 2, 384)) >= 0){
		/* Trying to remove prior semaphores. */
		if(semctl(semid, 0, IPC_RMID, 0) < 0){
			fprintf(stderr, "client:\tcould not remove semaphores "
					"with key 4754h\n");
			exit(1);
		}
	}
	if((semid = semget(0x4754, 2, IPC_CREAT | IPC_EXCL | 384)) < 0){
		fprintf(stderr, "client:\tcould not get semaphores "
				"with key 4754h\n");
		exit(1);
	}
	atexit(&client_atexit);
	/* do our own interrupt handling */
	for (i = 0; i < sizeof(sig2catch); i++) {
		siga.sa_handler = signal_handler;
		siga.sa_flags = 0;
		zero_sa_mask(&(siga.sa_mask));
		sigaction((int) sig2catch[i], &siga, old_signal_handler + i);
    	}
	if(!(child_id = vfork())){
		if(execl("/usr/bin/vgaserver", 0)){
			fprintf(stderr, "client: could not find"
				" server.\n");
			exit(1);
		}
	}
	/* Attaching shared memory segment */
	if((input_space = shmat(shmid, 0, 0)) == -1){
		fprintf(stderr, "client:\tcould not attach shared memory "
				"segment\n");
		exit(1);
	}
	output_space = input_space + 4096;
	((int *) output_space)[0] = mode;
	((int *) output_space)[1] = flags;
	wake_up_server();
	svgalib_linear_mem_size = ((int *) input_space)[2];
	svgalib_linear_mem_base = (char *) ((int *) input_space)[1];
	if(!svgalib_linear_mem_size){
		fprintf(stderr, "client: linear memory map not available\n");
		exit(1);
	}
	client_memfd = open("/dev/svga", O_RDWR);
	if(client_memfd < 0){
		fprintf(stderr, "client: cannot open /dev/svga\n");
		exit(1);
	}
	if((client_linear_mem_virtual = mmap((caddr_t) 0,
					svgalib_linear_mem_size,
					PROT_WRITE |  PROT_READ, 
					MAP_SHARED,
					client_memfd,
					(off_t)svgalib_linear_mem_base)) < 0){
		fprintf(stderr, "client: mmap failed.\n");
		exit(1);
	}
	graph_mem = client_linear_mem_virtual;
	return ((int *) input_space)[0]; 
}
int vga_init(void)
{
	return client_init(0, 0);
}

int vga_initf(int flags)
{
	return client_init(1, flags);
}


unsigned char *vga_getgraphmem(void)
{
	return graph_mem;
}

CODE_0_4(vga_setlogicalwidth, 15);

void vga_setdisplaystart(int a)
{
	((int *) output_space)[0] = 16;
	((int *) output_space)[1] = a;
	wake_up_server();
	client_displaystart = a;
}

CODE_0_0(vga_waitretrace, 17);
CODE_4_4(vga_claimvideomemory, 18);
CODE_0_0(vga_disabledriverreport, 19);
CODE_4_0(vga_getmonitortype, 20);
CODE_0_0(vga_lockvc, 21);
CODE_0_0(vga_unlockvc, 22);
int vga_simple_init(void)
{
	return client_init(2, 0);
}
CODE_4_0(vga_white, 23);
int vga_setrgbcolor(int r, int g, int b)
{
	
	if(client_modeinfo . bytesperpixel == 2){
		client_color = ((r >> 3) << 11) | ((g >> 2) << 5) | (b >> 3);
	} else {
		client_color = r << 16 | g << 8 | b;
	}
}
int vga_setlinearaddressing(void)
{
	return 0;
}

void vga_safety_fork(void (*f)(void))
{
	/* It is not easy to determine how to do a safety fork in
	   a client-server; so I just switched it off. --Mihai */
}

void vga_gettextfont(void *font)
{
	((int *) output_space)[0] = 24;
	wake_up_server();
	memcpy(font, input_space, FONT_SIZE);
}

void vga_puttextfont(void *font)
{
	((int *) output_space)[0] = 25;
	memcpy(output_space, font, FONT_SIZE);
	wake_up_server();
}

int vga_getcurrenttiming(int *x0, int *x1, int *x2, int *x3, int *x4, int *x5,
	int *x6, int *x7, int *x8, int *x9)
{
	((int *) output_space)[0] = 26;
	wake_up_server();
	*x0 = ((int *) input_space)[0];
	*x1 = ((int *) input_space)[1];
	*x2 = ((int *) input_space)[2];
	*x3 = ((int *) input_space)[3];
	*x4 = ((int *) input_space)[4];
	*x5 = ((int *) input_space)[5];
	*x6 = ((int *) input_space)[6];
	*x7 = ((int *) input_space)[7];
	*x8 = ((int *) input_space)[8];
	*x9 = ((int *) input_space)[9];
}

