#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <vga.h>
#include "libvga.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include <sys/types.h>

#define VIDEO
/*
#define HAS_CARD_INFO
*/
static unsigned char *input_space, *output_space;

static int shmid, semid;
static struct sembuf semarray[2];

void server_setmode(int mode)
{
	*((int *) output_space) = vga_setmode(mode);
	*((vga_modeinfo *) (output_space + sizeof(int))) = 
		*vga_getmodeinfo(mode);
}

void server_getpalette(void)
{
	*output_space = vga_getpalette(*input_space, ((int *)output_space) + 1, 
			((int *)output_space) + 2, ((int *) output_space) + 3);
}

void server_setpalvec(void)
{
	*output_space = vga_setpalvec(*((int *)input_space), 
			*(((int *)input_space) + 1), ((int *)input_space) + 2);
}

void server_getpalvec(void)
{
	*((int *)output_space) = vga_getpalvec(* ((int *)input_space), 
			*(((int *)input_space) + 1), 
			((int *)output_space) + 1);
}
void server_getmodeinfo(void)
{
	*((vga_modeinfo *) output_space) = *vga_getmodeinfo(
			((int *) input_space)[0]);
}

#if defined(HAS_CARD_INFO)

void server_getcardinfo(void)
{
	*((vga_cardinfo *) output_space) = *vga_getcardinfo();
}

#endif

void server_getmodename(void)
{
	strcpy(output_space, vga_getmodename(((int *) input_space)[0]));
}

void server_gettextfont(void)
{
	vga_gettextfont(output_space);
}

void server_puttextfont(void)
{
	vga_puttextfont(input_space);
}

void server_getcurrenttiming(void)
{
	vga_getcurrenttiming((int *) output_space,
			(int *)(output_space + 1 * sizeof(int)),
			(int *)(output_space + 2 * sizeof(int)),
			(int *)(output_space + 3 * sizeof(int)),
			(int *)(output_space + 4 * sizeof(int)),
			(int *)(output_space + 5 * sizeof(int)),
			(int *)(output_space + 6 * sizeof(int)),
			(int *)(output_space + 7 * sizeof(int)),
			(int *)(output_space + 8 * sizeof(int)),
			(int *)(output_space + 9 * sizeof(int)));
}

void server_terminate(void)
{
	if(shmdt(input_space))
		fprintf(stderr, "server: unable to detach shared memory"
			       " segment\n");
	/* Wake up client */
	semarray[0] . sem_num = 0;
	semarray[0] . sem_op = 1;
	semarray[0] . sem_flg = 0;
	if(semop(semid, semarray, 1)){
		fprintf(stderr, "server: could not perform semaphore "
				"operation\n");
		exit(1);
	}
	exit(0);
}

int vga_version = 0;		/* No version information for now. --Mihai */	

struct request_t{
	int normal;
       	void *function;
	int bytes_in;
}handlers[48] =	{
	{1, &server_setmode, 4},		/*0*/
	{1, &vga_hasmode, 4},
	{1, &vga_flip, 0},
	{1, &vga_setpalette, 16},		
	{0, &server_getpalette, 16},
	{0, &server_setpalvec, 1032},		/*5*/
	{0, &server_getpalvec, 8},
	{1, &vga_screenoff, 0},	
	{1, &vga_screenon, 0},	
	{1, &vga_dumpregs, 0},
	/* The functions setcolor, drawpixel, drawline, drawscanline and
	   drawscansegment are handled locally by the client */
	{0, &server_getmodeinfo, 4},		/*10*/
	{1, &vga_getdefaultmode, 0},
	{0, &server_getmodename, 4},
	{1, &vga_getmodenumber, 4},
	{1, &vga_getoptmode, 20},
	{1, &vga_setlogicalwidth, 4},		/*15*/
	{1, &vga_setdisplaystart, 4},
	{1, &vga_waitretrace, 0},
	{1, &vga_claimvideomemory, 4},
	{1, &vga_disabledriverreport, 0},
	{1, &vga_getmonitortype, 0},		/*20*/
	{1, &vga_lockvc, 0},	
	{1, &vga_unlockvc, 0},	
	{1, &vga_white, 0},
	/* vga_setegacolor unsupported */
	{0, &server_gettextfont, 0},		
	{0, &server_puttextfont, 0},		/*25*/
	{0, &server_getcurrenttiming, 0},
#if defined(HAS_CARD_INFO)
	{0, &server_getcardinfo, 0},
#endif
	{0, &server_terminate, 0}
};	
 
int main(int argc, char **argv)
{
	int (*function)(void);
	/* Attaching a shared memory segment */	
	if((shmid = shmget(0x4754, 8000, 384)) < 0){
		fprintf(stderr, "server:\tcould not create shared memory"
				"segment\n\twith key 4754h.\n");
		exit(1);
	}
	if((output_space = shmat(shmid, 0, 0)) == -1){
		fprintf(stderr, "server:\tcould not attach shared memory "
				"segment\n");
		exit(1);
	}
	input_space = output_space + 4096;
	/* Getting two semaphores for process synchonization */
	if((semid = semget(0x4754, 2, 384)) < 0){
		fprintf(stderr, "server:\tcould not access semaphores"
				"with key 4754h\n");
		exit(1);
	}
	/* Awaiting that the client gives us the initialization
	   information.
	 */
	semarray[0] . sem_num = 1;
	semarray[0] . sem_op = -1;
	semarray[0] . sem_flg = 0;
	if(semop(semid, semarray, 1)){
		fprintf(stderr, "server: could not perform semaphore "
				"operation\n");
		exit(1);
	}
	if(!((int *) input_space)[0])
		((int *) output_space)[0] = vga_init();
	else if(((int *) input_space)[0] == 1)
		((int *) output_space)[0] = vga_initf(((int *) input_space)[1]);
	else ((int *) output_space)[0] = vga_simple_init();
	((int *) output_space)[1] = __svgalib_linear_mem_base; 
	((int *) output_space)[2] = __svgalib_linear_mem_size;
	for(;;){
		semarray[0] . sem_num = 0;
		semarray[0] . sem_op = 1;
		semarray[0] . sem_flg = 0;
		semarray[1] . sem_num = 1;
		semarray[1] . sem_op = -1;
		semarray[1] . sem_flg = 0;
		if(semop(semid, semarray, 1)){
			fprintf(stderr, "server: could not perform semaphore "
					"operation\n");
			exit(1);
		}
		if(semop(semid, semarray + 1, 1)){
			fprintf(stderr, "server: could not perform semaphore "
					"operation\n");
			exit(1);
		}
#if defined(VIDEO)	
		function = handlers[((int *)input_space)[0]].function;
		if(handlers[((int *)input_space)[0]] . normal){
			switch(handlers[((int *)input_space)[0]] . bytes_in){
			case 0: *((int *)output_space) = 
					(*function)();
				break;
			case 4: *((int *)output_space) = 
					(*(int (*)(int))function)
					( ((int *)input_space)[1]);
				break;
			case 16: *((int *)output_space) = 
					(*(int (*)(int, int, int, int))function)
					( ((int *)input_space)[1],
					((int *)input_space)[2],
					((int *)input_space)[3],
					((int *)input_space)[4]);
				break;
			case 20: *((int *)output_space) = 
					(*(int (*)(int, int, int, int, int))
					 function) ( ((int *)input_space)[1],
					((int *)input_space)[2],
					((int *)input_space)[3],
					((int *)input_space)[4],
					((int *) input_space)[5]);
				break;
			default:	fprintf(stderr, 
						"internal server error\n");
					exit(1);
			}
		} else if(function) {
			(*function)();
		}
#endif
	}
}

